# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"extract_patches": "core.ipynb",
         "build_image_pyramid": "core.ipynb",
         "extract_patches_Ms": "core.ipynb",
         "convert_cv2_keypoints": "core.ipynb",
         "convert_cv2_plus_A_keypoints": "core.ipynb",
         "convert_xyA": "core.ipynb",
         "convert_LAFs": "laf.ipynb",
         "Ell2LAF": "laf.ipynb",
         "invSqrt": "laf.ipynb",
         "rectifyAffineTransformationUpIsUp": "laf.ipynb",
         "convert_ellipse_keypoints": "laf.ipynb",
         "ells2LAFs": "laf.ipynb",
         "LAF2pts": "laf.ipynb",
         "convertLAFs_to_A23format": "laf.ipynb",
         "LAFs2ell": "laf.ipynb",
         "visualize_LAFs": "laf.ipynb"}

modules = ["core.py",
           "laf.py"]

doc_url = "https://ducha-aiki.github.io/extract_patches/"

git_url = "https://github.com/ducha-aiki/extract_patches/tree/master/"

def custom_doc_links(name): return None